#!/usr/bin/ruby

require 'optparse'
require 'intelhex'
require 'ndump'

#	AKI-PIC Programmer Ver.4				   __U7__
#					~Pin Diagram~	Vpp 2-1  1-|	|-40 DATA
#											 2-|	|-39 CLK
#	JP2 2-1: for 40/28 pin device			 3-|	|-38
#	JP2 2-3: for 8 pin device				 4-|	|-37
#											 5-|	|-36 GND
#				   __U6__					 6-|	|-35
#			DATA 1-|	|-28				 7-|	|-34
#			Vdd	 2-|	|-27			GND	 8-|	|-33
#				 3-|	|-26				 9-|	|-32 Vdd
#			GND	 4-|	|-25				10-|	|-31 GND
#				 5-|	|-24			Vdd	11-|	|-30
#				 6-|	|-23			GND	12-|	|-29
#				 7-|	|-22				13-|	|-28
#				 8-|	|-21				14-|	|-27
#			Vpp	 9-|	|-20				15-|	|-26
#				10-|	|-19 Vdd			16-|	|-25
#				11-|	|-18 DATA		Vdd	17-|	|-24 GND
#				12-|	|-17 CLK			18-|	|-23 DATA
#				13-|	|-16				19-|	|-22 CLK
#				14-|____|-15		Vpp 2-3 20-|____|-21

class PIC

	attr_accessor	:configs
	attr_reader		:device_name
	attr_reader		:package
	attr_reader		:jumper2
	attr_reader		:program_memory
	attr_reader		:data_memory
	attr_reader		:config_words
	attr_reader		:device_id
	attr_reader		:device_mask

	def initialize(writer)
		@writer			= writer
		@configs		= Hash.new
		@device_name	= self.class
		@device_mask	= 0b11_1111_1110_0000
	end

	def send_writer_command(command, parameter)
		command.each_byte {|c|
			@writer.putc(c)
			sleep 0.001
		}
		parameter.each {|c|
			@writer.putc(c)
			sleep 0.001
		}
	end

	def check_writer_firmware
		send_writer_command('***?', [])
		firm_version = ''
		(0...17).each {|n|
			firm_version << @writer.getc.chr
		}
		@writer.getc == ?@ ? firm_version : false
	end

	def check_type(objects, address = 0x2006 * 2)
		object = objects[address & 0xFF00]
		device_id = (object[address & 0xFF] << 8) + object[(address + 1) & 0xFF]
		[device_id & @device_mask == @device_id,				# ident
			device_id & @device_mask,							# device id
			device_id & ~@device_mask,							# revision
		]
	end

	def get_config_words(objects, addresses = [0x2007 * 2])
		configs = Array.new
		addresses.each {|address|
			object = objects[address & 0xFF00]
			config_words = (object[address & 0xFF] << 8) + object[(address + 1) & 0xFF]
			config_types = @config_types[address]
			config_types.keys.sort {|a, b|						# マスクビット上位順に並べ替え
				(config_types[a][0] <=> config_types[b][0]) * -1
			}.each {|field|
				config_types[field][1].each {|state, value|
					configs << field << state if(value == config_words & config_types[field][0])
				}
			}
		}
		@configs = Hash[*configs]
		configs													# 表示用
	end

	def set_config_words(objects, addresses = [0x2007 * 2])
	end

	def read_general(object, words, dump = nil)
		dump ||= Ndump.new
		(0...words).each {|p|
			object << @writer.getc; yield(p, dump << object.last)
			object << @writer.getc; yield(p, dump << object.last)
			@writer.putc(?@) if(p % 8 == 7)
		}
		yield(words, dump.tail)
	end

	def write_general(object, words, dump = nil)
		dump ||= Ndump.new
		(0...words).each {|p|
			@writer.getc
			@writer.putc(object[p * 2    ]); yield(p, dump << object[p * 2    ])
			@writer.putc(object[p * 2 + 1]); yield(p, dump << object[p * 2 + 1])
		}
		yield(words, dump.tail)
	end

	def check_response
		@writer.getc == ?@ ? true : false
	end
end

#-------------------------------------------------------------------------------
#
#	PIC12F629 / 675
#
class PIC12F629 < PIC

	def initialize(writer)
		super
		@package		= 8						# 8pin
		@jumper2		= '2-3'
		@program_memory	= 1024					# words
		@data_memory	= 128					# bytes
		@config_words	= [0b11_0001_1111_1111]
		@device_id		= 0b00_1111_1000_0000

		@config_types	= Hash.new
			types = Hash.new
			types['BG'] = [				0b11_0000_0000_0000,	Hash[
				'LOWEST',				0b11_0000_0000_0000,
				'HIGHEST',				0b00_0000_0000_0000,
			]]
			types['CPD'] = [			0b00_0001_0000_0000, 	Hash[
				'DISABLE',				0b00_0001_0000_0000,
				'ENABLE',				0b00_0000_0000_0000,
			]]
			types['CP'] = [				0b00_0000_1000_0000, 	Hash[
				'DISABLE',				0b00_0000_1000_0000,
				'ENABLE',				0b00_0000_0000_0000,
			]]
			types['BODEN'] = [			0b00_0000_0100_0000, 	Hash[
				'ENABLE',				0b00_0000_0100_0000,
				'DISABLE',				0b00_0000_0000_0000,
			]]
			types['MCLRE'] = [			0b00_0000_0010_0000, 	Hash[
				'MCLR',					0b00_0000_0010_0000,
				'GP3',					0b00_0000_0000_0000,
			]]
			types['PWRTE'] = [			0b00_0000_0001_0000, 	Hash[
				'DISABLE',				0b00_0000_0001_0000,
				'ENABLE',				0b00_0000_0000_0000,
			]]
			types['WDTE'] = [			0b00_0000_0000_1000, 	Hash[
				'ENABLE',				0b00_0000_0000_1000,
				'DISABLE',				0b00_0000_0000_0000,
			]]
			types['FOSC'] = [			0b00_0000_0000_0111, 	Hash[
				'RC:GP4=CLKOUT',		0b00_0000_0000_0111,
				'RC:GP4=I/O',			0b00_0000_0000_0110,
				'INTOSC:GP4=CLKOUT',	0b00_0000_0000_0101,
				'INTOSC:GP4=I/O',		0b00_0000_0000_0100,
				'EC',					0b00_0000_0000_0011,
				'HS',					0b00_0000_0000_0010,
				'XT',					0b00_0000_0000_0001,
				'LP',					0b00_0000_0000_0000,
			]]
		@config_types[0x2007 * 2] = types
	end

	def erase
		send_writer_command('nbd', [7])
		check_response
	end

	def read_program(objects, address = 0x0000 * 2)
		send_writer_command('nrsp', [@program_memory >> 8, @program_memory & 0xFF])
		read_general(objects[address] = Array.new, @program_memory, Ndump.new(address, 'skip2')) {|progress, report|
			yield(progress, report)
		}
		check_response
	end

	def read_config(objects, address = 0x2000 * 2)
		send_writer_command('nrsc', [0, 8])
		read_general(objects[address] = Array.new, 8, Ndump.new(address, 'skip2')) {|progress, report|
			yield(progress, report)
		}
		check_response
	end

	def read_data(objects, address = 0x2100 * 2)
		send_writer_command('nrsd', [@data_memory >> 8, @data_memory & 0xFF])
		read_general(objects[address] = Array.new, @data_memory, Ndump.new(address)) {|progress, report|
			yield(progress, report)
		}
		check_response
	end

	def write_program(objects, address = 0x0000 * 2)
		object = objects[address]
		send_writer_command('nwee', [object.size >> 9, object.size >> 1 & 0xFF])
		write_general(object, object.size >> 1, Ndump.new(address, 'skip2')) {|progress, report|
			yield(progress, report)
		}
		check_response
	end

	def write_config(objects, address = 0x2000 * 2)
		object = objects[address]
		send_writer_command('nwec', [0, 8])
		write_general(object, object.size >> 1, Ndump.new(address, 'skip2')) {|progress, report|
			yield(progress, report)
		}
		check_response
	end

	def write_data(objects, address = 0x2100 * 2)
		object = objects[address]
		send_writer_command('nwed', [object.size >> 9, object.size >> 1 & 0xFF])
		write_general(object, object.size >> 1, Ndump.new(address)) {|progress, report|
			yield(progress, report)
		}
		check_response
	end
end

class PIC12F675 < PIC12F629

	def initialize(writer)
		super
		@device_id		= 0b00_1111_1100_0000
	end
end

#-------------------------------------------------------------------------------
#
#	PIC16F627A / 628A / 648A
#
class PIC16F627A < PIC

	def initialize(writer)
		super
		@package		= 18					# 18pin
		@jumper2		= 'NON'
		@program_memory	= 1024					# words
		@data_memory	= 128					# bytes
		@config_words	= [0b10_0001_1111_1111]
		@device_id		= 0b01_0000_0100_0000

		@config_types	= Hash.new
			types = Hash.new
			types['CP'] = [				0b10_0000_0000_0000, 	Hash[
				'DISABLE',				0b10_0000_0000_0000,
				'ENABLE',				0b00_0000_0000_0000,
			]]
			types['CPD'] = [			0b00_0001_0000_0000, 	Hash[
				'DISABLE',				0b00_0001_0000_0000,
				'ENABLE',				0b00_0000_0000_0000,
			]]
			types['LVP'] = [			0b00_0000_1000_0000, 	Hash[
				'ENABLE',				0b00_0000_1000_0000,
				'DISABLE',				0b00_0000_0000_0000,
			]]
			types['BODEN'] = [			0b00_0000_0100_0000, 	Hash[
				'ENABLE',				0b00_0000_0100_0000,
				'DISABLE',				0b00_0000_0000_0000,
			]]
			types['MCLRE'] = [			0b00_0000_0010_0000, 	Hash[
				'MCLR',					0b00_0000_0010_0000,
				'RA5',					0b00_0000_0000_0000,
			]]
			types['PWRTE'] = [			0b00_0000_0000_1000, 	Hash[
				'DISABLE',				0b00_0000_0000_1000,
				'ENABLE',				0b00_0000_0000_0000,
			]]
			types['WDTE'] = [			0b00_0000_0000_0100, 	Hash[
				'ENABLE',				0b00_0000_0000_0100,
				'DISABLE',				0b00_0000_0000_0000,
			]]
			types['FOSC'] = [			0b00_0000_0001_0011, 	Hash[
				'RC:RA6=CLKOUT',		0b00_0000_0001_0011,
				'RC:RA6=I/O',			0b00_0000_0001_0010,
				'INTOSC:RA6=CLKOUT',	0b00_0000_0001_0001,
				'INTOSC:RA6=I/O',		0b00_0000_0001_0000,
				'EC',					0b00_0000_0000_0011,
				'HS',					0b00_0000_0000_0010,
				'XT',					0b00_0000_0000_0001,
				'LP',					0b00_0000_0000_0000,
			]]
		@config_types[0x2007 * 2] = types
	end

	def erase
		send_writer_command('nbd', [7])
		check_response
	end

	def read_program(objects)
		send_writer_command('nrsp', [@program_memory >> 8, @program_memory & 0xFF])
		read_general(objects[0x0000 * 2] = Array.new, @program_memory) {|progress, report|
			yield(progress, report)
		}
		check_response
	end

	def read_config(objects)
		send_writer_command('nrsc', [0, 8])
		read_general(objects[0x2000 * 2] = Array.new, 8) {|progress, report|
			yield(progress, report)
		}
		check_response
	end

	def read_data(objects)
		send_writer_command('nrsd', [@data_memory >> 8, @data_memory & 0xFF])
		read_general(objects[0x2100 * 2] = Array.new, @data_memory) {|progress, report|
			yield(progress, report)
		}
		check_response
	end

	def write_program(objects)
		object = objects[0x0000 * 2]
		send_writer_command('nwe8', [object.size >> 9, object.size >> 1 & 0xFF])
		write_general(object, object.size >> 1) {|progress, report|
			yield(progress, report)
		}
		check_response
	end

	def write_config(objects)
		object = objects[0x2000 * 2]
		send_writer_command('nwec', [0, 8])
		write_general(object, object.size >> 1) {|progress, report|
			yield(progress, report)
		}
		check_response
	end

	def write_data(objects)
		object = objects[0x2100 * 2]
		send_writer_command('nwed', [object.size >> 9, object.size >> 1 & 0xFF])
		write_general(object, object.size >> 1) {|progress, report|
			yield(progress, report)
		}
		check_response
	end
end

class PIC16F628A < PIC16F627A
	def initialize(writer)
		super
		@program_memory	= 2048					# words
		@device_id		= 0b01_0000_0110_0000
	end
end

class PIC16F648A < PIC16F627A
	def initialize(writer)
		super
		@program_memory	= 4096					# words
		@data_memory	= 256					# bytes
		@device_id		= 0b01_0001_0000_0000
	end
end

#-------------------------------------------------------------------------------
#
#	PIC16F818 / 819
#
class PIC16F818 < PIC

	def initialize(writer)
		super
		@package		= 18					# 18pin
		@jumper2		= 'NON'
		@program_memory	= 1024					# words
		@data_memory	= 128					# bytes
		@config_words	= [0b11_1111_1111_1111]
		@device_id		= 0b00_0100_1100_0000
		@device_mask	= 0b11_1111_1111_0000

		@config_types	= Hash.new
			types = Hash.new
			types['CP'] = [				0b10_0000_1000_0000, 	Hash[
				'DISABLE',				0b10_0000_1000_0000,
				'ENABLE',				0b00_0000_0000_0000,
			]]
			types['CCPMX'] = [			0b01_0000_0000_0000, 	Hash[
				'RB2',					0b01_0000_0000_0000,
				'RB3',					0b00_0000_0000_0000,
			]]
			types['DEBUG'] = [			0b00_1000_0000_0000, 	Hash[
				'DISABLE',				0b00_1000_0000_0000,
				'ENABLE',				0b00_0000_0000_0000,
			]]
			types['WRT'] = [			0b00_0110_0000_0000, 	Hash[
				'DISABLE',				0b00_0110_0000_0000,
				'ENABLE',				0b00_0000_0000_0000,
			]]
			types['CPD'] = [			0b00_0001_0000_0000, 	Hash[
				'DISABLE',				0b00_0001_0000_0000,
				'ENABLE',				0b00_0000_0000_0000,
			]]
			types['LVP'] = [			0b00_0000_1000_0000, 	Hash[
				'ENABLE',				0b00_0000_1000_0000,
				'DISABLE',				0b00_0000_0000_0000,
			]]
			types['BODEN'] = [			0b00_0000_0100_0000, 	Hash[
				'ENABLE',				0b00_0000_0100_0000,
				'DISABLE',				0b00_0000_0000_0000,
			]]
			types['MCLRE'] = [			0b00_0000_0010_0000, 	Hash[
				'MCLR',					0b00_0000_0010_0000,
				'RA5',					0b00_0000_0000_0000,
			]]
			types['PWRTEN'] = [			0b00_0000_0000_1000, 	Hash[
				'DISABLE',				0b00_0000_0000_1000,
				'ENABLE',				0b00_0000_0000_0000,
			]]
			types['WDTEN'] = [			0b00_0000_0000_0100, 	Hash[
				'ENABLE',				0b00_0000_0000_0100,
				'DISABLE',				0b00_0000_0000_0000,
			]]
			types['FOSC'] = [			0b00_0000_0001_0011, 	Hash[
				'RC:RA6=CLKOUT',		0b00_0000_0001_0011,
				'RC:RA6=I/O',			0b00_0000_0001_0010,
				'INTOSC:RA6=CLKOUT',	0b00_0000_0001_0001,
				'INTOSC:RA6=I/O',		0b00_0000_0001_0000,
				'EC',					0b00_0000_0000_0011,
				'HS',					0b00_0000_0000_0010,
				'XT',					0b00_0000_0000_0001,
				'LP',					0b00_0000_0000_0000,
			]]
		@config_types[0x2007 * 2] = types
	end

	def erase
		send_writer_command('Nbd', [10])
		check_response
	end

	def read_program(objects)
		send_writer_command('Nrsp', [@program_memory >> 8, @program_memory & 0xFF])
		read_general(objects[0x0000 * 2] = Array.new, @program_memory) {|progress, report|
			yield(progress, report)
		}
		check_response
	end

	def read_config(objects)
		send_writer_command('Nrsc', [0, 8])
		read_general(objects[0x2000 * 2] = Array.new, 8) {|progress, report|
			yield(progress, report)
		}
		check_response
	end

	def read_data(objects)
		send_writer_command('Nrsd', [@data_memory >> 8, @data_memory & 0xFF])
		read_general(objects[0x2100 * 2] = Array.new, @data_memory) {|progress, report|
			yield(progress, report)
		}
		check_response
	end

	def write_program(objects)
		object = objects[0x0000 * 2]
		send_writer_command('NwHe', [object.size >> 9, object.size >> 1 & 0xFF])
		write_general(object, object.size >> 1) {|progress, report|
			yield(progress, report)
		}
		check_response
	end

	def write_config(objects)
		object = objects[0x2000 * 2]
		send_writer_command('NwFc', [0, 8])
		write_general(object, object.size >> 1) {|progress, report|
			yield(progress, report)
		}
		check_response
	end

	def write_data(objects)
		object = objects[0x2100 * 2]
		send_writer_command('NwId', [object.size >> 9, object.size >> 1 & 0xFF])
		write_general(object, object.size >> 1) {|progress, report|
			yield(progress, report)
		}
		check_response
	end
end

class PIC16F819 < PIC16F818

	def initialize(writer)
		super
		@program_memory	= 2048					# words
		@data_memory	= 256					# bytes
		@device_id		= 0b00_0100_1110_0000
	end
end

#-------------------------------------------------------------------------------
#
#	PIC16F882 / 883 / 884 / 886 / 887
#
class PIC16F882 < PIC

	def initialize(writer)
		super
		@package		= 28					# 28pin
		@jumper2		= '2-1'
		@program_memory	= 2048					# words
		@data_memory	= 128					# bytes
		@config_words	= [0b0011_1111_1111_1111, 0b0000_0111_0000_0000]
		@device_id		= 0b10_0000_0000_0000

		@config_types	= Hash.new
			types = Hash.new
			types['DEBUG'] = [			0b10_0000_0000_0000, 	Hash[
				'DISABLE',				0b10_0000_0000_0000,
				'ENABLE',				0b00_0000_0000_0000,
			]]
			types['LVP'] = [			0b01_0000_0000_0000, 	Hash[
				'ENABLE',				0b01_0000_0000_0000,
				'DISABLE',				0b00_0000_0000_0000,
			]]
			types['FCMEN'] = [			0b00_1000_0000_0000, 	Hash[
				'ENABLE',				0b00_1000_0000_0000,
				'DISABLE',				0b00_0000_0000_0000,
			]]
			types['IESO'] = [			0b00_0100_0000_0000, 	Hash[
				'ENABLE',				0b00_0100_0000_0000,
				'DISABLE',				0b00_0000_0000_0000,
			]]
			types['BOREN'] = [			0b00_0011_0000_0000, 	Hash[
				'ENABLE',				0b00_0011_0000_0000,
				'ENABLE w/o SLEEP',		0b00_0010_0000_0000,
				'ENABLE with SBOREN',	0b00_0001_0000_0000,
				'DISABLE',				0b00_0000_0000_0000,
			]]
			types['CPD'] = [			0b00_0000_1000_0000, 	Hash[
				'DISABLE',				0b00_0000_1000_0000,
				'ENABLE',				0b00_0000_0000_0000,
			]]
			types['CP'] = [				0b00_0000_0100_0000, 	Hash[
				'DISABLE',				0b00_0000_0100_0000,
				'ENABLE',				0b00_0000_0000_0000,
			]]
			types['MCLRE'] = [			0b00_0000_0010_0000, 	Hash[
				'MCLR',					0b00_0000_0010_0000,
				'RE3',					0b00_0000_0000_0000,
			]]
			types['PWRTE'] = [			0b00_0000_0001_0000, 	Hash[
				'DISABLE',				0b00_0000_0001_0000,
				'ENABLE',				0b00_0000_0000_0000,
			]]
			types['WDTEN'] = [			0b00_0000_0000_1000, 	Hash[
				'ENABLE',				0b00_0000_0000_1000,
				'DISABLE',				0b00_0000_0000_0000,
			]]
			types['FOSC'] = [			0b00_0000_0001_0011, 	Hash[
				'RC:RA6=CLKOUT',		0b00_0000_0001_0011,
				'RC:RA6=I/O',			0b00_0000_0001_0010,
				'INTOSC:RA6=CLKOUT',	0b00_0000_0001_0001,
				'INTOSC:RA6=I/O',		0b00_0000_0001_0000,
				'EC',					0b00_0000_0000_0011,
				'HS',					0b00_0000_0000_0010,
				'XT',					0b00_0000_0000_0001,
				'LP',					0b00_0000_0000_0000,
			]]
		@config_types[0x2007 * 2] = types
			types = Hash.new
			types['WRT'] = [			0b00_0110_0000_0000, 	Hash[
				'DISABLE',				0b00_0110_0000_0000,
				'ENABLE',				0b00_0000_0000_0000,
			]]
			types['BOR4V'] = [			0b00_0001_0000_0000, 	Hash[
				'4V',					0b00_0001_0000_0000,
				'2.1V',					0b00_0000_0000_0000,
			]]
		@config_types[0x2008 * 2] = types
	end

	def get_config_words(objects, addresses = [0x2007 * 2]) #, 0x2008 * 2])
		super
	end

	def erase
		send_writer_command('nbd', [7])
		check_response
	end

	def read_program(objects)
		send_writer_command('nrsp', [@program_memory >> 8, @program_memory & 0xFF])
		read_general(objects[0x0000 * 2] = Array.new, @program_memory) {|progress, report|
			yield(progress, report)
		}
		check_response
	end

	def read_config(objects)
		send_writer_command('nrsc', [0, 8])
		read_general(objects[0x2000 * 2] = Array.new, 8) {|progress, report|
			yield(progress, report)
		}
		check_response
	end

	def read_data(objects)
		send_writer_command('nrsd', [@data_memory >> 8, @data_memory & 0xFF])
		read_general(objects[0x2100 * 2] = Array.new, @data_memory) {|progress, report|
			yield(progress, report)
		}
		check_response
	end
end

class PIC16F883 < PIC16F882

	def initialize(writer)
		super
		@program_memory	= 4096					# words
		@data_memory	= 256					# bytes
		@device_id		= 0b10_0000_0010_0000
	end
end

class PIC16F884 < PIC16F883

	def initialize(writer)
		super
		@package		= 40					# 40pin
		@device_id		= 0b10_0000_0100_0000
	end
end

class PIC16F886 < PIC16F883

	def initialize(writer)
		super
		@program_memory	= 8192					# words
		@device_id		= 0b10_0000_0110_0000
	end
end

class PIC16F887 < PIC16F884

	def initialize(writer)
		super
		@program_memory	= 8192					# words
		@device_id		= 0b10_0000_1000_0000
	end
end

#-------------------------------------------------------------------------------
#
#	Main
#
device = nil
do_erase = nil
do_write = nil
do_read = nil

opts = OptionParser.new
opts.version = '0.1'
opts.release = '1'
opts.on('-d', '--device TYPE',	'set device type.')			{|p| device = p.upcase }
opts.on('-e', '--erase',		'erase program and data.')	{|p| do_erase = true }
opts.on('-w', '--write',		'write program.')			{|p| do_write = true }
opts.on('-r', '--read',			'read program.')			{|p| do_read  = true }
hexfile = opts.permute(ARGV)[0]

program = IntelHex.new
program.decode_hex(File.open(hexfile)) if(hexfile)

open('/dev/ttyUSB0', 'w+') {|writer|

	begin
		pic = eval(device + '.new(writer)')
	rescue
		raise "Device [#{device}] unknown."
	end

	firmware = pic.check_writer_firmware.chomp
	raise 'Failed to identify writer firmware.' unless(firmware)
	raise "Writer firmware [#{firmware}] not support." unless(firmware >= 'AE-PGM877  V6.71')
	print "==== Writer version: #{firmware} ===\n"

	pic.read_config(program.object)			{|progress, report| print report }
	print "\n"

	identify = pic.check_type(program.object)
	print "==== Device: #{device}  "\
		"Id: #{sprintf('%04X', identify[1])}...#{identify[0] ? 'ok' : '*** differ ***'}  "\
		"Rev: #{identify[2]} ====\n\n"
	raise 'Device mismatch.' unless(identify[0])

	configs = pic.get_config_words(program.object)
	print "==== Configuration word ====\n"
	(0...configs.size >> 1).each {|n|
		printf("%8s: %s\n", configs[n * 2], configs[n * 2 + 1]) 
	}

	if(do_erase or do_write)
		print "==== Erasing... ====\n"
		p pic.erase
	end

	if(do_write)
		print "==== Writing... ====\n"
		pic.write_program(program.object)	{|progress, report| print report }
		pic.write_data(program.object)		{|progress, report| print report }
	end

	if(do_read)
		print "==== Reading... ====\n"
		pic.read_program(program.object)	{|progress, report| print report }
		pic.read_data(program.object)		{|progress, report| print report }
	end
}

if(hexfile)
	program.encode_hex
	program.intelhex.each {|l|
		print "#{l}\n"
	}
end

__END__

